/* form-get.c -- Retrieve a particular form.

	Copyright 1994,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the package's "LICENSE" file for terms.


   Invoker passes form information in the PATH_INFO, as:

	directory/formname:path-tag


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>

#include "mvform.h"

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-cgi.h>

#include "formmisc.h"


/* Local Definitions */


/* External data referenced */

extern	BSTR	*GbsP;			/* General BSTR */
extern	BSTR	*GtkbsP;		/* BSTR for tokens */
extern	TKNS	*GtknsP;		/* Tokens pool */

/* External routines used */

extern	char	*getenv();

/* Local data publicly available */

	char	*DocumentRootP;		/* Document root */
	char	Formdir[PATH_MAX +1];	/* Dir for this form */

/* Local routines and forward references */



/* Private data */

static	char	*Form_filenameP;
static	char	*Path_infoP;
static	char	*Path_info_nameP;
static	char	Path_info_required = FALSE;
static	char	*Path_tagP;

main( argc, argv )
	int		argc;
	char		**argv;
{
	char		*tagP;
	char		*errP;
	char		*formnameP;	/* name of form */
	char		*cP, *c1P;	/* Character ptrs */
	void		*qP;

    if ( !formmisc_init() )
	exit( 0 );

    /* Collect arguments */
    if ( ( errP = qargs_in( (char *)NULL ) ) != NULL ) {
	report_error( errP );
	exit( 0 );
    }

    cP = getenv( "PATH_INFO" );
    if ( ( cP != NULL ) && ( *cP == '/' ) )
	++cP;			/* Skip '/' */
    if ( ( cP == NULL ) || ( *cP == NUL ) ) {
	report_error( "no form name was supplied." );
	exit( 0 );
    }

    /* The path name looks like this:

	pathname/formname:form-argument

       pathname is relative to DOCUMENT_ROOT and is required;
       formname is required;
       form-argument is any optional info for the form.
    */
    DocumentRootP = getenv( "DOCUMENT_ROOT" );
    if ( ( DocumentRootP == NULL ) || ( *DocumentRootP != '/' ) ) {
	report_error( "Can't determine DOCUMENT_ROOT" );
	exit( 0 );
    }

    c1P = strrchr( cP, ':' );
    if ( c1P != NULL ) {
	*c1P++ = NUL;
	Path_infoP = c1P;
    }
    else
	Path_infoP = NULL;

    c1P = strrchr( cP, '/' );
    if ( c1P == NULL ) {
	report_error( "No directory specified for form." );
	exit( 0 );
    }
    *c1P++ = NUL;
    formnameP = c1P;
    if ( snprintf( &Formdir[0], sizeof(Formdir),
		   "%s/%s", DocumentRootP, cP ) >= sizeof(Formdir) ) {
	report_error( "Path too large: \"%s/%s\"", DocumentRootP, cP );
	exit( 0 );
    }

    /* Get form definitions */
    if ( ! get_form_defs( formnameP ) )
	exit( 0 );

    /* If path info required, make sure it is present. */
    if ( Path_info_required && ( Path_infoP == NULL ) ) {
	report_error( "No path_info given as required by form." );
	exit( 0 );
    }
    
    /* Return the form. */
    begin_doc_out();
    deliver_form( formnameP );

    exit( 0 );
}
/*

*//* get_form_defs( fornmameP )

	Scan the form-defs file

Accepts :

	formnameP	Name of form

Returns :

	< value >	TRUE if OK
			FALSE if not (with error already generated h)

*/

int
get_form_defs ARGLIST( (formnameP) )
    FARG( char		*formnameP )	/* Name of form */
{
	int		status;
	char		*strP;
	FILE		*fP;
	char		fdfname[PATH_MAX +1];	/* Name of formdef file */

    /* Try to open the definition file */
    if ( snprintf( &fdfname[0], sizeof(fdfname),
		   "%s/%s.%s",
		   Formdir, formnameP, FORMDEF ) >= sizeof(fdfname) ) {
	report_error( "Path to definition file too large: \"%s/%s.%s\"",
		     Formdir, formnameP, FORMDEF );
	return ( FALSE );
    }
    fP = fopen( &fdfname[0], "r" );
    if ( fP == NULL ) {
	report_error( "Can't find definition for form \"%s\"", formnameP );
	return ( FALSE );
    }

    /* Read and parse the defs */
    for ( status = TRUE; 
          status && ( getformline( fP, GbsP ) >= 0 );
	     ) {
	tkline( GbsP, GtknsP );

	if ( GtknsP->tp_tC == 0 )	/* Blank line */
	    continue;

	/* Switch on first token. */
	strP = (char *)TKN_P( GtknsP, 0 );
	// if ( strcmp( strP, "#" ) == 0 )
	if ( *strP == '#' )
	    continue;			/* Comment */

	else if ( stricmp( strP, "Form-file" ) == 0 ) {
	    Form_filenameP = newstr( "get_form_defs", 
				    "form filename", (char *)TKN_P( GtknsP, 1 ) );
	    if ( strchr( Form_filenameP, '/' ) != NULL ) {
		report_error( "'/' not allowed in form-file name." );
		return ( FALSE );
	    }
	}
	else if ( stricmp( strP, "Path-required" ) == 0 ) {
	    Path_info_required = TRUE;
	}
	else if ( stricmp( strP, "Path-tag" ) == 0 ) {
	    Path_tagP = newstr( "get_form_defs",
			       "path tag", (char *)TKN_P( GtknsP, 1 ) );

	    /* Enter the tag and value into the environment */
	    setenv( Path_tagP, Path_infoP, TRUE );
	}
	else {
	    report_error( "Bad token \"%s\" in form definition file.",
				strP );
	    status = FALSE;
	}
    }

    fclose( fP );
    return ( status );
}
/*

*//* deliver_form( formnameP )

	Send the form to stdout

Accepts :

	formnameP		Name of form

Returns :

	< nothing >

*/

int
deliver_form ARGLIST( (formnameP) )
    FARG( char		*formnameP )	/* Name of form */
{
	char		formfile[PATH_MAX +1];	/* Name of form file */

    /* Get name of form file */
    if ( Form_filenameP == NULL ) {
	report_error( "No form filename given." );
	return ( FALSE );
    }
    if ( snprintf( &formfile[0], sizeof(formfile),
		   "%s/%s", Formdir, Form_filenameP ) >= sizeof(formfile) ) {
	report_error( "Path to form file too large: \"%s/%s\"",
		     Formdir, Form_filenameP );
	return ( FALSE );
    }

    /* Use formfilecat to cat it to stdout */
    return ( formfilecat( stdout, &formfile[0] ) );
}
