/* formmisc.c -- Misc routines in support of form handling

	Copyright 1994,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the package's "LICENSE" file for terms.

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <sys/types.h>
#include <time.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>

#include "mvform.h"

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-cgi.h>

#include "formmisc.h"


/* Local Definitions */

#define	PROCCHAR	'~'		/* Char that indicates processing */
#define	VARMAX		100		/* Max size of input variable */
#define	NO_TRANSLATE	0x100		/* Set for any character that
					   should not be subject to
					   further expansion (e.g. from
					   a variable substitution.) */

    /* Number of strings in the safe-string ring */
#define	SAFESTRINGMAX	10

typedef enum {				/* Where input is coming from */
    IN_FILE,				/* From a file */
    IN_LITERAL,				/* From a literal string */
}  INPTYPE;

typedef struct inpctl {			/* Input control at a level */
    struct inpctl *ic_prevP;		/* Ptr to previous control */
    INPTYPE	ic_type;		/* Input type */
    void	*ic_ptr;		/* Any pointer associated.. */
    UBYTE	ic_ungetc;		/* Any character put back */
}  INPCTL;


typedef struct {			/* "safe" string buffer */
    char	*ss_strP;		/* Ptr to string buffer */
    int		ss_strL;		/* Length of string */
}  SAFESTR;


/* External data referenced */

extern	char	*DocumentRootP;		/* Name of document root */
extern	char	Formdir[];		/* Name of form directory */

/* External routines used */

extern	char	*getenv();

/* Local data publicly available */

	BOOL	Doc_out_begunF = FALSE;	/* Whether document output has been
					   started (via begin_doc_out)
					*/
	char	*Doc_out_mimeP;		/* Mime type for output doc */

	BSTR	*GbsP;			/* General BSTR */
	BSTR	*GtkbsP;		/* BSTR for tokens */
	TKNS	*GtknsP;		/* Tokens pool */
	BOOL	Report_errorF = FALSE;	/* If any report_error has been done */

/* Local routines and forward references */

static	void	getcondition PROTO( (FILE *fP) );
static	int	getformc PROTO( (FILE *fP) );
static	int	getinpc PROTO( (FILE *fP) );
static	void	getinpfname PROTO( (char *fnameP, int fnbuflen, FILE *fP) );
static	int	getinpnonblank PROTO( (FILE *fP) );
static	void	getinprelation PROTO( (char *relP, int relbuflen, FILE *fP) );
static	char	*getinpterm PROTO( (char *termP, int termbuflen, FILE *fP) );
static	char	*getinptoken PROTO( (char *tokP, int tokbuflen, FILE *fP) );
static	char	*getinpvar PROTO( (char *varbufP, int varbuflen, FILE *fP) );
static	char	*safestr PROTO( (char *strP) );
static	void	skipfalse PROTO( (FILE *fP) );
static	void	ungetinpc PROTO( (int ch, FILE *fP) );
static	void	pushinp PROTO( (INPTYPE inptype, void *inpP) );
static	void	popinp PROTO( (void) );

/* Private data */

static	INPCTL	*InplistP;		/* Input control list */
static	SAFESTR	Safestrings[SAFESTRINGMAX];
static	int	SafestrX;		/* Index into safestring */
static	int	TruthC;			/* Truth block count */

/*

*//* begin_doc_out()

	Begin document output to stdout

Accepts :


Returns :

	<nothing>

Notes :

	The output to stdout has to start with an HTTP prolog, which
	itself ends with a blank line.

	This function will be a no-op if document output has already
	been started (known via Doc_out_begunF).

	Otherwise, this function:

	 - emits the mime-type HTTP header field;
	 - emits a blank line.

	Note that if any other HTTP header fields are desired by the
	  caller, they can be output (put out?) before calling this
	  function.

*/

void
begin_doc_out NOARGLIST
{
    if ( Doc_out_begunF )
	return;				/* Already begun, do nothing */

    /* Add mime type field */
    printf( "Content-Type: %s\n",
	     Doc_out_mimeP == NULL ? "text/html" : Doc_out_mimeP );

    /* End the HTTP header */
    printf( "\n" );

    /* Make sure httpd header gets out */
    fflush( stdout );

    /* Remember that we did it */
    Doc_out_begunF = TRUE;

    /* done */
}
/*

*//* formfilecat( ofP, fnameP )

	Send a document to a file

Accepts :

	ofP	FILE ptr for output
	fnameP	Name of the file

Returns :

	< nothing >

*/

int
formfilecat ARGLIST( (ofP, fnameP) )
   NFARG( FILE		*ofP )		/* Output file ptr */
    FARG( char		*fnameP )	/* Name of form */
{
	BOOL		okF;
	FILE		*fP;
	char		linebuf[2000];

    /* Open the file */
    fP = fopen( fnameP, "r" );
    if ( fP == NULL ) {
	report_error( "formfilecat: can't find file \"%s\"", fnameP );
	return ( FALSE );
    }

    /* Copy the file linewise, with variable substitution */
    okF = TRUE;
    while ( okF && getformline( fP, GbsP ) >= 0 ) {
	if ( fprintf( ofP, "%s\n", (char *)GbsP->bs_bP ) <= 0 ) {
	    report_error( "formfilecat: write error %d", errno );
	    okF = FALSE;
	}
    }

    if ( fclose( fP ) != 0 ) {
	if ( okF ) {
	    report_error( "formfilecat: read/close error %d", errno );
	    okF = FALSE;
	}
    }

    return ( okF );
}
/*

*//* formmisc_init()

	Initialize some stuff that formmisc uses.

Accepts :

	<nothing>

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Sets up some stuff that is used by formmisc.
	Initializes the generic buffers and token pools that
	  utilities can make use of.

*/

BOOL
formmisc_init NOARGLIST
{
    /* Set up some buffers */
    GbsP = bstr_new( "general bstr", BIGBUFSIZE, BIGBUFSIZE );
    if ( GbsP == NULL ) {
	report_error( "Can't allocate GbsP" );
	return ( FALSE );
    }
    GtkbsP = bstr_new( "tokens bstr", BIGBUFSIZE, BIGBUFSIZE );
    if ( GtkbsP == NULL ) {
	report_error( "Can't allocate GtkbsP" );
	return ( FALSE );
    }
    GtknsP = tkns_new( GtkbsP, 50, 
			(UBYTE *)"", (UBYTE *)" \t,;:",
			FALSE );
    if ( GtknsP == NULL ) {
	report_error( "Can't allocate tokens pool" );
	return ( FALSE );
    }

    return ( TRUE );
}
/*

*//* getformline( fP, bsP )

	Input a line of text for form processing

Accepts :

	fP		Stream for input file
	bsP		A BSTR to input into

Returns :

	< value >	Number of characters read (-1 if EOF)

Notes :

	Excess input (up to a newline) is ignored.
	variable substitution is performed as per getformc

	The number of characters read may be more than the number
	stored.

*/

int
getformline ARGLIST( ( fP, bsP ) )
    NFARG( FILE		*fP) 		/* Input file ptr */
     FARG( BSTR		*bsP )		/* BSTR for the input */
{
	int		cC;		/* Character count */
	int		ch;		/* Character */

    bsP->bs_bC = 0;
    getformc( NULL );

    for( cC = 0; ( ( ch = getformc( fP ) ) != '\n' ) && ( ch != EOF ); ++cC ) {
	bstr_putb( bsP, ch );
    }

    bstr_nulterm( bsP );

    if ( ( cC == 0 ) && ( ch == EOF ) )
	cC = -1;

    return( cC );
}
/*

*//* report_error( fmtP, args... )

	Report an error in HTML

Accepts :

	fmtP		printf-style format
	args...		printf arguments

Returns :

	<nothing>

Notes :


*/

void
report_error ARGLIST( ( fmtP, va_decl ) )
   NFARG( char		*fmtP )
    FARG( ... )
{
	va_list		ap;

    /* Start varargs */
    va_start( ap, fmtP );

    /* Make sure document output has started */
    begin_doc_out();

    /* Do the output, with optional intro if first time */
    if ( !Report_errorF ) {
	printf( "\n<p>An error occurred during processing:<br>\n" );
	Report_errorF = TRUE;
    }

    vprintf( fmtP, ap );
    printf( "\n" );

    va_end( ap );
}

/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/

/*

*//* getformc( fP )

	Get a character from input for form processing

Accepts :

	fP		stream, or NULL to reset scanning context

Returns :

	< value >	character, or EOF

Notes:

	Performs variable substitution:
		~var		replaced with value of form tag
		~$var		replaced with environment var
		~+name		replaced with contents of file name
		~~		replaced with ~

	Parsing state is global, so caller should completely consume
	  a variable before switching streams, or clear the variable
	  by passing an fP of NULL.

*/


static int
getformc ARGLIST( (fP) )
    FARG( FILE		*fP )		/* Stream ptr */
{
	int		ch;
	char		*val1P, *val2P;	/* Ptrs to variable values */
	char		var1[VARMAX +1]; /* Variable name */
	char		var2[VARMAX +2];
	char		filename[PATH_MAX +1];

    if ( fP == NULL ) {
	/* Reset translation state */
	/* (well, that's a legacy, lets do without) */

	return 0;
    }

    /* Get character */
    ch = getinpc( fP );

    /* If it's EOF, return it now */
    if ( ch == EOF )
	return ( ch );

    /* Likewise if it has NO_TRANSLATE on */
    if ( ( ch & NO_TRANSLATE ) == NO_TRANSLATE )
	return ( ch & 0xff );

    /* Check for substitution */
    if ( ch != PROCCHAR )
	return ( ch );

    /* Got the trigger character.  What we do now depends on the
	next character.
    */
    ch = getinpc( fP );

    /* If EOF, just return the procchar and catch EOF next time through */
    if ( ch == EOF )
	return ( PROCCHAR );

    ch &= 0xff;				/* Strip any quote bits */

    if ( ch == PROCCHAR )		/* Doubled up ? */
	return ( ch );			/* Yep, return one of them. */

    /* Check for variable start */
    if ( ( ch == '$' ) ||
         ( ch == ':' ) ||
         isalpha( ch ) ||
	 isdigit( ch ) ||
	 ( ch == '\'' ) || ( ch == '"' ) ) {
	ungetinpc( ch, fP );		/* Put back the start char */
	val1P = getinpvar( &var1[0], sizeof(var1), fP );
	if ( val1P == NULL ) {
	    /* Variable lookup fails; push back a safe copy of
		the variable name string.
	    */
	    pushinp( IN_LITERAL, safestr( &var1[0] ) );
	    return ( PROCCHAR );	/* Return the proc char.. */
	}

	/* Variable lookup succeeds; push its translation into the
	   input stream and recurse to fetch from it.
	*/
	pushinp( IN_LITERAL, val1P );
	return ( getformc( fP ) );
    }

    /* Dispatch on other characters following the procchar */
    switch ( ch ) {
	case '+':			/* File name for inclusion */
	    getinpfname( &var1[0], sizeof( var1 ), fP );
	    if ( var1[0] != NUL ) {
		if ( ( makefilename( &filename[0], sizeof(filename),
				      &var1[0] ) ) &&
		     ( ( fP = fopen( &filename[0], "r" ) ) != NULL ) ) {
		    pushinp( IN_FILE, fP );
		    return ( getformc( fP ) );
		}
		/* Bad name or bad open, re-insert the filename. */
		pushinp( IN_LITERAL, safestr( &var1[0] ) );
	    }
	    ch = getformc( fP );
	    break;

	case '?':			/* Conditional */
	    getcondition( fP );
	    ch = getformc( fP );
	    break;

	case '}':			/* End of conditional */
	    if ( TruthC > 0 ) {
		--TruthC;
		ch = getinpc( fP );
		if ( ch == '{' ) {	/* An "else" body */
		    skipfalse( fP );
		}
		else if ( ( ch & 0xff ) == '-' ) {
		    /* Eat any following newline */
		    ch = getinpc( fP );
		    if ( ( ch & 0xff ) != '\n' )
			ungetinpc( ch, fP );
		}
		else
		    ungetinpc( ch, fP );
		ch = getformc( fP );
	    }
	    else {
		ungetinpc( ch, fP );
		ch = PROCCHAR;
		pushinp( IN_LITERAL,
			"\n<p><b>Mismatched closing of body.</b><p>\n" );
	    }
	    break;

        default:			/* Nothing special */
	    /* Push back the character and return the procchar */
	    ungetinpc( ch, fP );
	    ch = PROCCHAR;
	    break;
    }

    /* Return whatever we have left */
    return ( ch );
}
/*

*//* getcondition( fP )

	Process a condition next in the input stream

Accepts :

	fP		Top-level file pointer

Returns :

	nothing

*/

static void
getcondition ARGLIST( (fP) )
    FARG( FILE		*fP )		/* Top file ptr */
{
	char		truth;
	int		ch, ch2;
	char		*val1P, *val2P;	/* Ptrs to variable values */
	char		var1[VARMAX +1]; /* Variable name */
	char		var2[VARMAX +2];
	char		relation[ 10 ];
	char		pushback[ VARMAX *3 ];

    /* Get first term */
    val1P = getinpterm( &var1[0], sizeof(var1), fP );
    if ( val1P == NULL ) {
	/* No term value, push back the string. */
	if ( var1[0] == NUL )
	    pushinp( IN_LITERAL, safestr( &var1[0] ) );
	return;
    }

    /* Get the relation */
    getinprelation( &relation[0], sizeof(relation), fP );
    if ( relation[0] == NUL ) {
	/* No relation, push back term1 and go back */
	pushinp( IN_LITERAL, safestr( &var1[0] ) );
	return;
    }

    /* Get second term. */
    val2P = getinpterm( &var2[0], sizeof(var2), fP );
    if ( val2P == NULL ) {
	/* No second term; push back first term and relation */
	pushinp( IN_LITERAL, safestr( &relation[0] ) );
	pushinp( IN_LITERAL, safestr( &var1[0] ) );
	return;
    }

    /* Make sure there's a body following. */
    ch = getinpnonblank( fP );
    ch2 = getinpc( fP );
    if ( ( ch != PROCCHAR ) || ( ch2 != '{' ) ) {
	/* No body: push back everything to here. */
	ungetinpc( ch, fP );

	if ( ch2 != EOF )
	    snprintf( &pushback[0], sizeof(pushback), "%c%s%s%s",
			ch2, &var1[0], &relation[0], &var2[0] );
	else
	    snprintf( &pushback[0], sizeof(pushback), "%s%s%s",
			&var1[0], &relation[0], &var2[0] );
	pushinp( IN_LITERAL, safestr( &pushback[0] ) );

	snprintf( &pushback[0], sizeof(pushback), "%c?", PROCCHAR );
	pushinp( IN_LITERAL, safestr( &pushback[0] ) );

	return;
    }

    /* Check for a "-" continuation character */
    ch = getinpc( fP );
    if ( ( ch & 0xff ) == '-' ) {
	/* Yes; eat any following newline. */
	ch = getinpc( fP );
	if ( ( ch & 0xff ) != '\n' )
	    ungetinpc( ch, fP );
    }
    else
	ungetinpc( ch, fP );

    /* Evaluate the condition... brute force, sorry. */
    if ( strcmp( relation, "==" ) == 0 )
	/* Equality */
	truth = ( strcmp( val1P, val2P ) == 0 );
    else if ( strcmp( relation, "@==" ) == 0 )
	/* Case-insensitive equality */
	truth = ( stricmp( val1P, val2P ) == 0 );
    else if ( strcmp( relation, ".==" ) == 0 )
	/* regexp equality */
	truth = strmatch( (UBYTE *)val2P, (UBYTE *)val1P, SM_REGEX, 0 );
    else if ( strcmp( relation, "@.==" ) == 0 )
	/* regexp equality, case insensitive */
	truth = strmatch( (UBYTE *)val2P, (UBYTE *)val1P,
			   SM_REGEX, MF_IGNCASE );
    else if ( strcmp( relation, "?==" ) == 0 )
	/* wildmatch equality */
	truth = strmatch( (UBYTE *)val2P, (UBYTE *)val1P, SM_WILD, 0 );
    else if ( strcmp( relation, "@?==" ) == 0 )
	/* wildmatch equality, case insensitive */
	truth = strmatch( (UBYTE *)val2P, (UBYTE *)val1P,
			   SM_WILD, MF_IGNCASE );

    /* Notted versions of the above */
    else if ( strcmp( relation, "!=" ) == 0 )
	/* Inequality */
	truth = ( strcmp( val1P, val2P ) != 0 );
    else if ( ( strcmp( relation, "!@=" ) == 0 ) ||
              ( strcmp( relation, "@!=" ) == 0 ) )	// legacy
	/* Case-insensitive inequality */
	truth = ( stricmp( val1P, val2P ) != 0 );
    else if ( strcmp( relation, "!.=" ) == 0 )
	/* regexp inequality */
	truth = !strmatch( (UBYTE *)val2P, (UBYTE *)val1P, SM_REGEX, 0 );
    else if ( strcmp( relation, "!@.=" ) == 0 )
	/* regexp inequality, case insensitive */
	truth = !strmatch( (UBYTE *)val2P, (UBYTE *)val1P,
			   SM_REGEX, MF_IGNCASE );
    else if ( strcmp( relation, "!?=" ) == 0 )
	/* wildmatch inequality */
	truth = !strmatch( (UBYTE *)val2P, (UBYTE *)val1P, SM_WILD, 0 );
    else if ( strcmp( relation, "!@?=" ) == 0 )
	/* wildmatch inequality, case insensitive */
	truth = !strmatch( (UBYTE *)val2P, (UBYTE *)val1P,
			   SM_WILD, MF_IGNCASE );

    else {
	/* Invalid relation.  Assume true, but also insert a message. */
	snprintf( &pushback[0], sizeof(pushback), "%c?%s%s%s %c{",
			PROCCHAR,
			&var1[0], &relation[0], &var2[0],
			PROCCHAR );
	pushinp( IN_LITERAL, safestr( &pushback[0] ) );

	snprintf( &pushback[0], sizeof(pushback),
				"<p><b>Invalid relation \"%s\":</b><br>\n",
				&relation[0] );
	pushinp( IN_LITERAL, safestr( &pushback[0] ) );

	truth = TRUE;
    }

    if ( truth )
	/* Condition evaluates true, increment truth count. */
	++TruthC;
    else {
	skipfalse( fP );
    }
}
/*

*//* skipfalse( fP )

	Skip to the end of a false condition body

Accepts :

	fP		Input stream at top level


Returns :

	nothing

*/

static void
skipfalse ARGLIST( (fP) )
    FARG( FILE		*fP )		/* File at top level */
{
	int		ch;
	int		falseC;		/* Nesting level */

    falseC = 1;

    while ( ( ch = getinpc( fP ) ) != EOF ) {
	if ( ch != PROCCHAR )
	    continue;
	ch = getinpc( fP );
	if ( ch == '}' ) {
	    /* End of condition. */
	    if ( --falseC == 0 )
		break;
	    /* Check for else side */
	    ch = getinpc( fP );
	    if ( ch == '{' )
		++falseC;
	    else
		ungetinpc( ch, fP );
	}
	else if ( ch == '{' )
	    ++falseC;
    }

    /* Check for a true (else) conditional */
    ch = getinpc( fP );
    if ( ch == '{' )
	++TruthC;
    else
	ungetinpc( ch, fP );

    /* Check for trailing "-" continuation character */
    ch = getinpc( fP );
    if ( ( ch & 0xff ) == '-' ) {
	/* Yes; eat any following newline. */
	ch = getinpc( fP );
	if ( ( ch & 0xff ) != '\n' )
	    ungetinpc( ch, fP );
    }
    else
	ungetinpc( ch, fP );
}
/*

*//* getinpvar( varbufP, varbuflen, fP )

	Get variable name from input stream, and translate it.

Accepts :

	varbufP		Where to store the variable name
	varbuflen	Size of varbufP's buffer
	fP		File PTR at top of input stream

Returns :

	< value >	Ptr to variable translation if valid variable;
			NULL if no substitution.

*/

static char *
getinpvar ARGLIST( (varbufP, varbuflen, fP) )
   NFARG( char		*varbufP )	/* Ptr to variable name buffer */
   NFARG( int		varbuflen )	/* Size of variable name buffer */
    FARG( FILE		*fP )		/* top-level input file */
{
	int		qch;		/* If quoted, this is the quote char */
	int		ch;
	int		varX;		/* Index to variable name */
	long		timenow;
	struct tm	*tmP;
	char		*vP;
	char		*strP;
	void		*qP;		/* Query arg ptr */
	char		valbuf[ BIGBUFSIZE +1];
	char		datefmt[ 200 ];
	char		datestr[ 200 ];

    /* Collect the variable name */
    varX = 0;
    --varbuflen;			/* Allow for trailing NUL */

    /* Check for quoted variable */
    if ( ( qch = getinpc( fP ) ) != EOF ) {
	qch &= 0xff;
	if ( ( qch != '\'' ) && ( qch != '"' ) ) {
	    /* Not a quote character, put it back */
	    ungetinpc( qch, fP );
	    qch = 0;
	}
    }

    while ( ( ch = getinpc( fP ) ) != EOF ) {
	ch &= 0xff;

	/* If quoted string, accept the end quote as termination. */
	if ( qch != 0 ) {
	    if ( ch == qch )
		break;
	    if ( varX < varbuflen )
		varbufP[ varX++ ] = ch;
	}
	else {
	    /* We'll take letters, digits, _, -, ., or
	       leading dollarsign (for environment variable )
	    */
	    if ( ( ( varX == 0 ) &&
	           ( ( ch == '$' ) || ( ch == ':' ) ) ) ||
		isalpha( ch ) ||
		isdigit( ch ) ||
		( ch == '-' ) ||
		( ch == '_' ) ||
		( ch == '.' ) ) {
		/* Store this character or lose it if too many */
		if ( varX < varbuflen )
	            varbufP[ varX++ ] = ch;
	    }
	    else {
		ungetinpc( ch, fP );
		break;
	    }
	}
    }

    varbufP[ varX ] = NUL;		/* NUL-terminate variable */

    /* get the translation */
    if ( varbufP[0] == '$' ) {
	/* Environment variable lookup. */
	if ( varbufP[1] == NUL )
	    strP = NULL;		/* No var name given. */
	else if ( strcmp( &varbufP[1], "DATE" ) == 0 ) {
	    /* Special processing for DATE; check for +"format" */
	    ch = getinpc( fP );
	    if ( ( ch & 0xff ) != '+' ) {
		/* No format given, use a default. */
		ungetinpc( ch, fP );
		vP = "%c";
	    }
	    else
		/* Format given, get it. */
		vP = getinptoken( &datefmt[0], sizeof(datefmt), fP );

	    /* Format the time now according to the format. */
	    time( &timenow );
	    tmP = localtime( &timenow );
	    strftime( &datestr[0], sizeof(datestr), vP, tmP );
	    strP = safestr( &datestr[0] );
	}
	else {
	    strP = getenv( &varbufP[1] );
	    if ( strP != NULL )
		strP = safestr( strP );
	}
    }
    else if ( varbufP[0] == ':' ) {
	/* yes/no for tagname present */
	if ( varbufP[1] == NUL )
	    qP = NULL;
	else
	    qP = qarg_find_tag( &varbufP[1] );
	if ( ( qP == NULL ) ||
	     isblankstr( qarg_value( qP ) ) )
	    strP = "no";
	else
	    strP = "yes";
    }
    else {
	/* Tag name lookup */
	if ( varbufP[0] == NUL )
	    strP = NULL;		/* No tag name given */
	else {
	    qP = qarg_find_tag( varbufP );
	    if ( qP == NULL )
	        strP = NULL;
	    else {
		vP = &valbuf[0];
		while ( qP != NULL ) {
		    if ( stricmp( qarg_tag( qP ), varbufP ) != 0 )
			break;
		    if ( vP != &valbuf[0] )
			*vP++ = ',';
		    strcpy( vP, qarg_value( qP ) );
		    vP += strlen( vP );
		    qP = qarg_next( qP );
		}
		*vP = NUL;
		strP = safestr( &valbuf[0] );
	    }
	}
    }

    return ( strP );
}
/*

*//* getinpfname( fnameP, fnbuflen, fP )

	Get filename from input stream

Accepts :

	fnameP		Where to store the file name
	fnbuflen	Size of fnameP's buffer
	fP		File PTR at top of input stream

Returns :

	none

*/

static void
getinpfname ARGLIST( (fnameP, fnbuflen, fP) )
   NFARG( char		*fnameP )	/* Ptr to filename buffer */
   NFARG( int		fnbuflen )	/* Size of filename buffer */
    FARG( FILE		*fP )		/* top-level input file */
{
	int		ch;
	int		fnX;		/* Index to variable name */

    /* Collect the file name */
    fnX = 0;
    --fnbuflen;			/* Allow for trailing NUL */

#if 0
    while ( ( ch = getinpc( fP ) ) != EOF ) {
#else
    while ( ( ch = getformc( fP ) ) != EOF ) {
#endif
	ch &= 0xff;

	/* Take anything reasonable. */
	if ( isalpha( ch ) ||
	     isdigit( ch ) ||
	     ( ch == '/' ) ||
	     ( ch == '-' ) ||
	     ( ch == '_' ) ||
	     ( ch == '.' ) ) {
	    /* Store this character or lose it if too many */
	    if ( fnX < fnbuflen )
	        fnameP[ fnX++ ] = ch;
	}
	else {
	    ungetinpc( ch, fP );
	    break;
	}
    }

    fnameP[ fnX ] = NUL;		/* NUL-terminate variable */

    return;
}
/*

*//* getinpterm( termP, termbuflen, fP )

	Get condition term from input

Accepts :

	termP		Where to store the term string
	termbuflen	Size of termP's buffer
	fP		File PTR at top of input stream

Returns :

	< value >	Ptr to the value of the term if valid
			NULL if no valid term or substitution

*/

static char *
getinpterm ARGLIST( (termP, termbuflen, fP) )
   NFARG( char		*termP )	/* Ptr to term name buffer */
   NFARG( int		termbuflen )	/* Size of term name buffer */
    FARG( FILE		*fP )		/* top-level input file */
{
	int		ch;

    /* Check for variable reference */
    ch = getinpnonblank( fP );
    if ( ch == PROCCHAR )
	return( getinpvar( termP, termbuflen, fP ) );
    else {
	ungetinpc( ch, fP );
	return ( getinptoken( termP, termbuflen, fP ) );
    }
}
/*

*//* getinprelation( relP, relbuflen, fP )

	Get condition relation from input

Accepts :

	relP		Where to store the relation string
	relbuflen	Size of relP's buffer
	fP		File PTR at top of input stream

Returns :

	nothing

*/

static void
getinprelation ARGLIST( (relP, relbuflen, fP) )
   NFARG( char		*relP )		/* Ptr to relation buffer */
   NFARG( int		relbuflen )	/* Size of relation buffer */
    FARG( FILE		*fP )		/* top-level input file */
{
	int		ch,ch2;
	int		relX;		/* Index to relation buffer */

    /* Collect the relation */
    relX = 0;
    --relbuflen;			/* Allow for trailing NUL */

    ch = getinpnonblank( fP );
    while ( ch != EOF ) {
	ch2 = ch & 0xff;

	/* Take the characters that can appear in a relation */
	if ( ( ch2 == '@' ) ||
	     ( ch2 == '!' ) ||
	     ( ch2 == '?' ) ||
	     ( ch2 == '.' ) ||
	     ( ch2 == '=' ) ||
	     ( ch2 == '<' ) ||
	     ( ch2 == '>' ) ) {
	    /* Store this character or lose it if too many */
	    if ( relX < relbuflen )
	        relP[ relX++ ] = ch2;
	}
	else
	    break;

	ch = getinpc( fP );
    }

    ungetinpc( ch, fP );
    relP[ relX ] = NUL;		/* NUL-terminate relation */
}
/*

*//* getinptoken( tokP, tokbuflen, fP )

	Get word or token from input streak

Accepts :

	tokP		Where to store the token string
	tokbuflen	Size of tokP's buffer
	fP		File PTR at top of input stream

Returns :

	< value >	The token's value

Notes :

	A quoted string may include quotes by doubling them up.
	The "value" of a quoted string is the text between the quotes.

*/

static char *
getinptoken ARGLIST( (tokP, tokbuflen, fP) )
   NFARG( char		*tokP )		/* Ptr to token buffer */
   NFARG( int		tokbuflen )	/* Size of token buffer */
    FARG( FILE		*fP )		/* top-level input file */
{
	int		ch;
	int		qc;		/* quotechar */
	int		tokX;		/* Index to token buffer */
	char		*valP;		/* Value to return */

    /* Collect the token */
    tokX = 0;
    --tokbuflen;			/* Allow for trailing NUL */

    ch = getinpnonblank( fP );
    qc = ch & 0xff;
    if ( ( qc == '"' ) || ( qc == '\'' ) ) {
	tokP[ tokX++ ] = qc;
	tokbuflen -= 2;			/* For quotes */
	while ( ( ch = getinpc( fP ) ) != EOF ) {
	    ch = ch & 0xff;
	    if ( ch == qc ) {
		/* Check for doubled-up quote */
		ch = getinpc( fP );
		if ( ( ch == EOF ) || ( ( ch & 0xff ) != qc ) ) {
		    /* end of quoted string, be done. */
		    ungetinpc( ch, fP );
		    ch = qc;
		    break;
		}
	    }
	    if ( tokX < tokbuflen )
		tokP[ tokX++ ] = ch;
	}
	/* Get value of string */
	tokP[tokX] = NUL;
	valP = safestr( &tokP[1] );

	/* Close off the string */
	tokP[tokX++] = qc;
	tokP[tokX] = NUL;
    }
    else {
	/* Non-quoted string. */
	while ( ch != EOF ) {
	    ch &= 0xff;

	    /* Take the characters that can appear in a token */
	    if ( isalpha( ch ) ||
		 isdigit( ch ) ||
		 ( ch == '_' ) ) {

		/* Store this character or lose it if too many */
		if ( tokX < tokbuflen )
	            tokP[ tokX++ ] = ch;
	    }
	    else
		break;

	    ch = getinpc( fP );
	}
	ungetinpc( ch, fP );
	tokP[ tokX ] = NUL;

	if ( tokP[0] == NUL )
	    valP = NULL;
	else
	    valP = tokP;
    }
    return ( valP );
}
/*

*//* getinpnonblank( fP )

	Get next non-blank character from input stream.

Accepts :

	fP		FILE ptr at top level

Returns :

	<value>		EOF if end of input
			non-EOF if normal character.

*/

static int
getinpnonblank ARGLIST( (fP) )
    FARG( FILE		*fP )		/* top-level input file ptr */
{
	int		ch, ch2;

    while ( ( ch = getinpc( fP ) ) != EOF ) {
	ch2 = ch & 0xff;
	if ( ( ch != ' ' ) &&
	     ( ch != '\t' ) &&
	     ( ch != '\r' ) &&
	     ( ch != '\n' ) )
	    break;
    }

    return ( ch );
}
/*

*//* getinpc( fP )

	Get character from input stream

Accepts :

	fP		FILE ptr at top level

Returns :

	<value>		EOF if end of input
			non-EOF if normal character.

Notes :

    The returned character (if not EOF) will have the NO_TRANSLATE
    bit set if it should not be subject to translation.

*/

static int
getinpc ARGLIST( (fP) )
    FARG( FILE		*fP )		/* top-level input file ptr */
{
	int		ch;		/* The character */

    if ( InplistP == NULL )
	/* No low-level input stream; get character from top level file */
	ch = getc( fP );
    else {
	/* Take character from low-level stream */
	if ( InplistP->ic_ungetc != NUL ) {
	    /* There's a pushed-back character, grab it. */
	    ch = InplistP->ic_ungetc;
	    InplistP->ic_ungetc = NUL;
	}
	else {
	    /* Process by input type */
	    switch( InplistP->ic_type ) {
	        case	IN_FILE:
		    ch = getc( (FILE *)InplistP->ic_ptr );
		    break;

		case	IN_LITERAL:	/* A string */
		    ch = *(UBYTE *)(InplistP->ic_ptr );
		    if ( ch == NUL )
			ch = EOF;
		    else
			/* Char is OK, advance the pointer */
			++(UBYTE *)InplistP->ic_ptr;
		    break;

		default:
		    report_error( "Internal error, bad INPTYPE %d", 
						InplistP->ic_type );
		    ch = EOF;
		    break;
	    }
	}

	/* If not EOF, apply any quoting necessary */
	if ( ch != EOF ) {
	    switch( InplistP->ic_type ) {
		case	IN_LITERAL:
		    ch |= NO_TRANSLATE;
		    break;
	    }
	}
	else {
	    /* Got EOF; pop the input stream and try again. */
	    popinp();
	    ch = getinpc( fP );
	}
    }

    /* Strip CR characters (note... doesn't apply to NO_TRANSLATE) */
    if ( ch == '\r' )
	return ( getinpc( fP ) );

    return ( ch );
}
/*

*//* ungetinpc( ch, fP )

	Unget (push back) a character into the current stream

Accepts :

	ch		The character to push back
	fP		Top-level input file ptr

Returns :

	nothing

Notes :

	Only one character may be pushed back at any level

*/

static void
ungetinpc ARGLIST( (ch, fP) )
   NFARG( int		ch )		/* Character to push back */
    FARG( FILE		*fP )		/* top-level input file */
{
    if ( ch != EOF ) {
	if ( InplistP != NULL )
	    /* There's a lower level input context, push back to it. */
	    InplistP->ic_ungetc = ch & 0xff;
	else
	    /* Push back to the file stream */
	    ungetc( ch, fP );
    }
}
/*

*//* pushinp( inptype, inpP )

	Push to a new input context

Accepts :

	inptype		Input type
	inpP		Any pointer associated with this input type

Returns :

	nothing

*/

static void
pushinp ARGLIST( (inptype, inpP) )
   NFARG( INPTYPE	inptype )	/* Input type */
    FARG( void		*inpP )		/* Any input pointer */
{
	INPCTL		*icP;		/* Input control block */

    icP = (INPCTL *)emalloc( "pushinp", "input block", sizeof(INPCTL) );

    icP->ic_type = inptype;
    icP->ic_ptr = inpP;
    icP->ic_ungetc = NUL;

    /* make this the top block */
    icP->ic_prevP = InplistP;
    InplistP = icP;
}
/*

*//* popinp()

	Pop the current input stream

Accepts :


Returns :


*/

static void
popinp NOARGLIST
{
	INPCTL		*icP;		/* Control block */

    icP = InplistP;
    if ( icP == NULL )
	return;

    /* Pop the input control off of the stack */
    InplistP = icP->ic_prevP;

    /* Do any type-specific cleanup */
    switch( icP->ic_type ) {
	case	IN_FILE:		/* File */
	    fclose( (FILE *)icP->ic_ptr );
	    break;

	default:			/* No action for other types */
	    break;
    }

    /* Delete the block */
    free( icP );
}
/*

*//* safestr( strP )

	Return a "safe" copy of a string

Accepts :

	strP		Ptr to string to save

Return :

	< value >	Ptr to safe copy of string


Notes :

	Uses a ring buffer of safe strings; the size of this buffer is
essentially the usage life of a safe string.

*/

static char *
safestr ARGLIST( (strP) )
    FARG( char		*strP )		/* String to save */
{
	SAFESTR		*ssP;		/* Safe string block */

    /* Get next safe string block */
    ssP = &Safestrings[ SafestrX++ ];
    if ( SafestrX == SAFESTRINGMAX )
	SafestrX = 0;

    /* Make sure it's got enough space */
    if ( ssP->ss_strP == NULL ) {
	ssP->ss_strL = strlen( strP ) + 20;
	ssP->ss_strP = emalloc( "safestr", "string buf", ssP->ss_strL );
    }
    else if ( ssP->ss_strL < ( strlen(strP) +1) ) {
	ssP->ss_strL = strlen( strP ) + 20;
	ssP->ss_strP = erealloc( "safestr", "string buf",
					ssP->ss_strP, ssP->ss_strL );
    }

    strcpy( ssP->ss_strP, strP );
    return ( ssP->ss_strP );
}
/*

*//* makefilename( dfnameP, dfnameL, sfnameP )

	Construct a filename from a filename reference.

Accepts :

	dfnameP		Where to store the filename
	dfnameL		Size of the dfnameP buffer
	sfnameP		Source (reference) filename

Returns :

	< value >	TRUE if OK
			FALSE if error

Notes :

	Makes sure that the resulting filename is in an acceptable
	directory.

*/

int
makefilename ARGLIST( (dfnameP, dfnameL, sfnameP) )
   NFARG( char		*dfnameP )	/* Destination filename */
   NFARG( int		dfnameL )	/* Length of dfnameP buffer */
    FARG( char		*sfnameP )	/* Source filename */
{
    /* Make sure no "../" in the path */
    if ( ( strncmp( sfnameP, "../", 3 ) == 0 ) ||
	 ( strstr( sfnameP, "/../" ) != NULL ) ) {
	report_error( "Invalid filename reference: %s", sfnameP );
	return( FALSE );
    }

    if ( *sfnameP == '/' ) {
	/* Absolute ref is taken relative to DOCUMENT_ROOT */
	if ( snprintf( dfnameP, dfnameL, "%s%s",
		       DocumentRootP, sfnameP ) >= dfnameL ) {
	    report_error( "Path too long: \"%s%s\"",
			  DocumentRootP, sfnameP );
	    return ( FALSE );
	}
    }
    else {
	/* Relative ref is with respect to formdir */
	if ( snprintf( dfnameP, dfnameL, "%s/%s",
		       &Formdir[0], sfnameP ) >= dfnameL ) {
	    report_error( "Path too long: \"%s/%s\"",
			 &Formdir[0], sfnameP );
	    return ( FALSE );
	}
    }

    return ( TRUE );
}
